@@ -1,5 +1,5 @@
 from tdw.FBOutput import Vector3, Quaternion, PassMask, Color, MessageType, MachineType, SimpleTransform, PathState
-from tdw.FBOutput import Environments as Envs
+from tdw.FBOutput import SceneRegions as SceRegs
 from tdw.FBOutput import Transforms as Trans
 from tdw.FBOutput import Rigidbodies as Rigis
 from tdw.FBOutput import Bounds as Bouns
@@ -44,6 +44,9 @@ from tdw.FBOutput import MagnebotWheels as MWheels
 from tdw.FBOutput import Occlusion as Occl
 from tdw.FBOutput import Lights as Lites
 from tdw.FBOutput import Categories as Cats
+from tdw.FBOutput import StaticRigidbodies as StatRig
+from tdw.FBOutput import RobotJointVelocities as RoJoVe
+from tdw.FBOutput import EmptyObjects as Empty
 import numpy as np
 from typing import Tuple, Optional
 
@@ -129,21 +132,21 @@ class OutputData(object):
         return color.R(), color.G(), color.B()
 
 
-class Environments(OutputData):
-    def get_data(self) -> Envs.Environments:
-        return Envs.Environments.GetRootAsEnvironments(self.bytes, 0)
+class SceneRegions(OutputData):
+    def get_data(self) -> SceRegs.SceneRegions:
+        return SceRegs.SceneRegions.GetRootAsSceneRegions(self.bytes, 0)
 
     def get_center(self, index: int) -> Tuple[float, float, float]:
-        return OutputData._get_vector3(self.data.Envs(index).Center)
+        return OutputData._get_vector3(self.data.Regions(index).Center)
 
     def get_bounds(self, index: int) -> Tuple[float, float, float]:
-        return OutputData._get_vector3(self.data.Envs(index).Bounds)
+        return OutputData._get_vector3(self.data.Regions(index).Bounds)
 
     def get_id(self, index: int) -> int:
-        return self.data.Envs(index).Id()
+        return self.data.Regions(index).Id()
 
     def get_num(self) -> int:
-        return self.data.EnvsLength()
+        return self.data.RegionsLength()
 
 
 class Transforms(OutputData):
@@ -182,6 +185,20 @@ class Rigidbodies(OutputData):
     def get_angular_velocity(self, index: int) -> Tuple[float, float, float]:
         return OutputData._get_vector3(self.data.Objects(index).AngularVelocity)
 
+    def get_sleeping(self, index: int) -> bool:
+        return self.data.Objects(index).Sleeping()
+
+
+class StaticRigidbodies(OutputData):
+    def get_data(self) -> StatRig.StaticRigidbodies:
+        return StatRig.StaticRigidbodies.GetRootAsStaticRigidbodies(self.bytes, 0)
+
+    def get_num(self) -> int:
+        return self.data.ObjectsLength()
+
+    def get_id(self, index: int) -> int:
+        return self.data.Objects(index).Id()
+
     def get_mass(self, index: int) -> float:
         return self.data.Objects(index).Mass()
 
@@ -191,6 +208,15 @@ class Rigidbodies(OutputData):
     def get_kinematic(self, index: int) -> bool:
         return self.data.Objects(index).Kinematic()
 
+    def get_dynamic_friction(self, index: int) -> float:
+        return self.data.Objects(index).DynamicFriction()
+
+    def get_static_friction(self, index: int) -> float:
+        return self.data.Objects(index).StaticFriction()
+
+    def get_bounciness(self, index: int) -> float:
+        return self.data.Objects(index).Bounciness()
+
 
 class Bounds(OutputData):
     def get_data(self) -> Bouns.Bounds:
@@ -504,6 +530,9 @@ class ImageSensors(OutputData):
     def get_sensor_forward(self, index: int) -> Tuple[float, float, float]:
         return OutputData._get_xyz(self.data.Sensors(index).Forward())
 
+    def get_sensor_field_of_view(self, index: int) -> float:
+        return self.data.Sensors(index).FieldOfView()
+
 
 class CameraMatrices(OutputData):
     def get_data(self) -> CaMa.CameraMatrices:
@@ -773,6 +802,9 @@ class AudioSources(OutputData):
     def get_is_playing(self, index: int) -> bool:
         return self.data.Objects(index).IsPlaying()
 
+    def get_samples(self) -> np.array:
+        return self.data.SamplesAsNumpy()
+
 
 class Raycast(OutputData):
     def get_data(self) -> Ray.Raycast:
@@ -943,6 +975,29 @@ class Robot(OutputData):
         return self.data.Immovable()
 
 
+class RobotJointVelocities(OutputData):
+    def get_data(self) -> RoJoVe.RobotJointVelocities:
+        return RoJoVe.RobotJointVelocities.GetRootAsRobotJointVelocities(self.bytes, 0)
+
+    def get_id(self) -> int:
+        return self.data.Id()
+
+    def get_num_joints(self) -> int:
+        return self.data.JointsLength()
+
+    def get_joint_id(self, index: int) -> int:
+        return self.data.Joints(index).Id()
+
+    def get_joint_velocity(self, index: int) -> np.array:
+        return self.data.Joints(index).VelocityAsNumpy()
+
+    def get_joint_angular_velocity(self, index: int) -> np.array:
+        return self.data.Joints(index).AngularVelocityAsNumpy()
+
+    def get_joint_sleeping(self, index: int) -> bool:
+        return self.data.Joints(index).Sleeping()
+
+
 class Keyboard(OutputData):
     def get_data(self) -> Key.Keyboard:
         return Key.Keyboard.GetRootAsKeyboard(self.bytes, 0)
@@ -1126,3 +1181,22 @@ class Categories(OutputData):
 
     def get_category_color(self, index: int) -> Tuple[float, float, float]:
         return OutputData._get_rgb(self.data.CategoryData(index).Color())
+
+
+class EmptyObjects(OutputData):
+    def __init__(self, b):
+        super().__init__(b)
+        self._ids = self.data.IdsAsNumpy().view(dtype=int)
+        self._positions = self.data.PositionsAsNumpy().view(dtype=np.float32).reshape(-1, 3)
+
+    def get_data(self) -> Empty.EmptyObjects:
+        return Empty.EmptyObjects.GetRootAsEmptyObjects(self.bytes, 0)
+
+    def get_num(self) -> int:
+        return len(self._ids)
+
+    def get_id(self, index: int) -> int:
+        return self._ids[index]
+
+    def get_position(self, index: int) -> np.array:
+        return self._positions[index]
