@@ -717,3 +717,13 @@ class TDWUtils:
 
         # Source: https://github.com/Unity-Technologies/URDF-Importer/blob/c41208565419b04907496baa93ad1b675d41dc20/com.unity.robotics.urdf-importer/Runtime/Extensions/TransformExtensions.cs#L85-L92
         return np.radians(np.array([-euler_angles[2], euler_angles[0], -euler_angles[1]]))
+
+    @staticmethod
+    def bytes_to_megabytes(b: int) -> float:
+        """
+        :param b: A quantity of bytes.
+
+        :return: A quantity of megabytes.
+        """
+
+        return b / (1 << 20)
